/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "Icons.h"

Icons::Icons(HWND parentWnd)
{
	_parentWnd = parentWnd;

	// Icon settings
	_iconList = NULL;
	_iconsLoaded = FALSE;

	_visible = GetRCBoolDef("DesktopIconVisible", TRUE);
	_iconX = GetRCInt("DesktopIconX", 0);
	_iconY = GetRCInt("DesktopIconY", 0);
	_iconSize = GetRCInt("DesktopIconSize", 32);
	_iconSpacingX = GetRCInt("DesktopIconSpacingX", 36);
	_iconSpacingY = GetRCInt("DesktopIconSpacingY", 42);

	GetRCString("DesktopIconFolder", _iconFolder, "C:\\", MAX_PATH);
	if(_iconFolder[strlen(_iconFolder)-1] != '\\') {
		strcat(_iconFolder, "\\");
	}

	_iconHideExtension = GetRCBoolDef("DesktopIconHideExtension", TRUE);
	_iconFoldersFirst = GetRCBoolDef("DesktopIconFoldersFirst", TRUE);
	_iconShowHidden = GetRCBoolDef("DesktopIconShowHidden", FALSE);

	char buffer[MAX_LINE_LENGTH];

	GetRCString("DesktopIconTextBackground", buffer, "", MAX_LINE_LENGTH);
	if(strlen(buffer)) {
		_iconTextBackground = CreateSolidBrush(GetRCColor("DesktopIconTextBackground", RGB(255,255,255)));
	}
	else {
		_iconTextBackground = NULL;
	}
	_iconTextSpacing = GetRCInt("DesktopIconTextSpacing", 5);
	_iconTextHeight = GetRCInt("DesktopIconTextHeight", 28);
	_iconTextWidth = GetRCInt("DesktopIconTextWidth", 64);

	_iconTextColor = GetRCColor("DesktopIconTextFontColor", RGB(0,0,0));

	GetRCString("DesktopIconTextFont", buffer, "Arial", MAX_LINE_LENGTH);
	int fontSize = GetRCInt("DesktopIconTextFontSize", 14);
	BOOL fontItalic = GetRCBoolDef("DesktopIconTextFontItalic", FALSE);
	BOOL fontBold = GetRCBoolDef("DesktopIconTextFontBold", FALSE);
	if(!(_iconTextFont = CreateFont(fontSize, 0, 0, 0, fontBold ? FW_BOLD : FW_NORMAL, fontItalic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, buffer))) {
		LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not load font \"%s\"", buffer);
	}

	GetRCString("DesktopIconDirection", buffer, "", MAX_LINE_LENGTH);
	if(!stricmp(buffer, "down")) {
		_iconDirection = DD_DOWN;
	}
	else if(!stricmp(buffer, "left")) {
		_iconDirection = DD_LEFT;
	}
	else if(!stricmp(buffer, "up")) {
		_iconDirection = DD_UP;
	}
	else {
		_iconDirection = DD_RIGHT;
	}

	GetRCString("DesktopIconWrapDirection", buffer, "", MAX_LINE_LENGTH);
	if(_iconDirection == DD_DOWN || _iconDirection == DD_UP) {
		if(!stricmp(buffer, "left")) {
			_iconWrapDirection = DD_LEFT;
		}
		else {
			_iconWrapDirection = DD_RIGHT;
		}
	}
	else {
		if(!stricmp(buffer, "up")) {
			_iconWrapDirection = DD_UP;
		}
		else {
			_iconWrapDirection = DD_DOWN;
		}
	}
	_iconWrapCount = max(GetRCInt("DesktopIconWrapCount", 10), 1);

	iconChangeHandle = FindFirstChangeNotification(_iconFolder, FALSE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME);
	if(iconChangeHandle == INVALID_HANDLE_VALUE) {
		LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not create change notification handle");
	}
}

Icons::~Icons()
{
	_deleteIcons();

	if(_iconTextFont) {
		DeleteObject(_iconTextFont);
	}
	if(_iconTextBackground) {
		DeleteObject(_iconTextBackground);
	}

	FindCloseChangeNotification(iconChangeHandle);
}

void Icons::show()
{
	_visible = TRUE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Icons::hide()
{
	_visible = FALSE;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Icons::toggle()
{
	_visible = !_visible;
	InvalidateRect(_parentWnd, NULL, TRUE);
}

void Icons::update()
{
	_deleteIcons();
	_loadIcons();
}

BOOL Icons::mouseEvent(Message& message)
{
	DESKTOPICON *clicked = _getIconAtPt(message.lParamLo, message.lParamHi);

	if(clicked) {
		switch(message.uMsg) {
			case WM_RBUTTONUP:
				PopupContextMenu(clicked->filename);
				break;

			case WM_LBUTTONDBLCLK:
				LSExecuteEx(NULL, NULL, clicked->filename, NULL, NULL, SW_NORMAL);
				break;

			case WM_LBUTTONDOWN:
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDOWN:
				if(_selectedIcon) {
					InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
				}
				_selectedIcon = clicked;
				if(_selectedIcon) {
					InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
				}
				break;
		}
		return TRUE;
	}
	else {
		if(_selectedIcon) {
			InvalidateRect(_parentWnd, &_selectedIcon->textRect, TRUE);
		}
		_selectedIcon = NULL;
		return FALSE;
	}
}

void Icons::_deleteIcons()
{
	DESKTOPICON *loop = _iconList;
	while(loop) {
		loop = _iconList->next;
		if(_iconList->icon) {
			DeleteObject(_iconList->icon);
		}
		free(_iconList);
		_iconList = loop;
	}
	_iconsLoaded = FALSE;
}

void Icons::_loadIcons()
{
	// Build file list
	WIN32_FIND_DATA fd;
	HANDLE search;
	char filter[MAX_PATH] = "\0";
	BOOL finished = FALSE;

	// Create file filter
	strcpy(filter, _iconFolder);
	strcat(filter, "*");

	// Start the search
	search = FindFirstFile(filter, &fd);
	if(search == INVALID_HANDLE_VALUE) {
		LSLogPrintf(LOG_DEBUG, "IconDesktop", "No files found.");
	}

	// Loop through the files
	while(!finished) {
		// Don't include "." and ".."
		if(strcmp(fd.cFileName, ".") && strcmp(fd.cFileName, "..")) {
			// Create a icon object
			DESKTOPICON *icon = (DESKTOPICON *)malloc(sizeof(DESKTOPICON));
			memset(icon, 0, sizeof(DESKTOPICON));

			// Add it to the list
			if(_iconList) {
				// Find last item in list
				DESKTOPICON *last = _iconList;
				while(last->next) {
					last = last->next;
				}
				last->next = icon;
			}
			else {
				_iconList = icon;
			}

			// Save file info
			strcpy(icon->filename, _iconFolder);
			strcat(icon->filename, fd.cFileName);
			strcpy(icon->title, fd.cFileName);
			if(_iconHideExtension) {
				PathRemoveExtension(icon->title);
			}
		}

		// Find next file
		if(!FindNextFile(search, &fd)) {
			if(GetLastError() == ERROR_NO_MORE_FILES) {
				finished = TRUE;
			}
			else {
				LSLogPrintf(LOG_ERROR, "IconDesktop", "Couldn't find next file.");
			}
		}
	}

	// Close the search handle.
	if(!FindClose(search)) {
		LSLogPrintf(LOG_WARNING, "IconDesktop", "Couldn't close search handle.");
	}

	// Hide system files and hidden files
	if(!_iconShowHidden) {
		DESKTOPICON *icon = NULL;
		DESKTOPICON *last = NULL;
		DESKTOPICON *loop = _iconList;
		_iconList = NULL;
		while(loop) {
			icon = loop;
			loop = loop->next;
			icon->next = NULL;

			DWORD attrib = GetFileAttributes(icon->filename);
			if((attrib & FILE_ATTRIBUTE_HIDDEN) != FILE_ATTRIBUTE_HIDDEN && (attrib & FILE_ATTRIBUTE_SYSTEM) != FILE_ATTRIBUTE_SYSTEM) {
				// Add it to the list
				if(_iconList) {
					// Find last item in list
					last = _iconList;
					while(last->next) {
						last = last->next;
					}
					last->next = icon;
				}
				else {
					_iconList = icon;
				}
			}
			else {
				delete icon;
			}
		}
	}

	// Folders first
	if(_iconFoldersFirst) {
		DESKTOPICON *loop = _iconList;
		DESKTOPICON *icon = NULL;
		DESKTOPICON *folders = NULL;
		DESKTOPICON *files = NULL;
		DESKTOPICON *last = NULL;
		DESKTOPICON **list = NULL;
		while(loop) {
			icon = loop;
			loop = loop->next;
			icon->next = NULL;

			if((GetFileAttributes(icon->filename) & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY) {
				list = &folders;
			}
			else {
				list = &files;
			}

			// Add it to the list
			if(*list) {
				// Find last item in list
				last = *list;
				while(last->next) {
					last = last->next;
				}
				last->next = icon;
			}
			else {
				*list = icon;
			}
		}

		// Merge lists
		if(folders) {
			_iconList = folders;
			if(files) {
				last = folders;
				while(last->next) {
					last = last->next;
				}
				last->next = files;
			}
		}
		else if(files) {
			_iconList = files;
		}
		else {
			_iconList = NULL;
		}
	}

	// Calculate RECT's and load icons
	int i = 0, x = 0, y = 0;
	DESKTOPICON *icon = _iconList;
	while(icon) {
		switch(_iconDirection) {
			case DD_RIGHT:
			case DD_LEFT:
				x = (i % _iconWrapCount);
				y = (i / _iconWrapCount);
				if(_iconWrapDirection == DD_UP) {
					y = -y;
				}
				if(_iconDirection == DD_LEFT) {
					x = -x;
				}
				break;

			case DD_DOWN:
			case DD_UP:
				x = (i / _iconWrapCount);
				y = (i % _iconWrapCount);
				if(_iconWrapDirection == DD_LEFT) {
					x = -x;
				}
				if(_iconDirection == DD_UP) {
					y = -y;
				}
				break;

		}

		icon->x = _iconX + ((_iconSize + _iconSpacingX) * x);
		icon->y = _iconY + ((_iconSize + _iconSpacingY) * y);

		icon->textRect.top = icon->y + _iconSize + _iconTextSpacing;
		icon->textRect.bottom = icon->textRect.top + _iconTextHeight;
		icon->textRect.left = icon->x + (_iconSize / 2) - (_iconTextWidth / 2);
		icon->textRect.right = icon->textRect.left + _iconTextWidth;

		SetRect(&icon->iconRect, icon->x, icon->y, icon->x + _iconSize, icon->y + _iconSize);

		icon->icon = _getIconFromPathEx(icon->filename, _iconSize, _iconSize);
		if(!icon->icon) {
			LSLogPrintf(LOG_ERROR, "IconDesktop", "Could not load icon for \"%s\"", icon->filename);
		}
		icon = icon->next;
		i++;
	}

	_iconsLoaded = TRUE;
}

void Icons::paintIcons(HDC hDC)
{
	if(_visible) {
		if(_iconsLoaded == FALSE) {
			_loadIcons();
		}

		COLORREF oldColor = SetTextColor(hDC, _iconTextColor);
		HFONT oldFont = (HFONT)SelectObject(hDC, _iconTextFont);
		int oldMode = SetBkMode(hDC, TRANSPARENT);

		DESKTOPICON *loop = _iconList;
		while(loop) {
			if(loop->icon) {
				DrawIconEx(hDC, loop->x, loop->y, loop->icon, _iconSize, _iconSize, 0, NULL, DI_NORMAL);
				if(_iconTextBackground) {
					FillRect(hDC, &loop->textRect, _iconTextBackground);
				}
				if(_selectedIcon == loop) {
					DrawFocusRect(hDC, &loop->textRect);
				}
				DrawText(hDC, loop->title, strlen(loop->title), &loop->textRect, DT_WORDBREAK | DT_CENTER | DT_NOPREFIX /*| DT_NOCLIP*/ | DT_END_ELLIPSIS);
			}
			loop = loop->next;
		}

		SetTextColor(hDC, oldColor);
		SelectObject(hDC, oldFont);
		SetBkMode(hDC, oldMode);
	}
}

DESKTOPICON *Icons::_getIconAtPt(int x, int y)
{
	POINT mouse = {x, y};
	DESKTOPICON *loop = _iconList;
	while(loop) {
		if(PtInRect(&loop->textRect, mouse) || PtInRect(&loop->iconRect, mouse)) {
			return loop;
		}
		loop = loop->next;
	}
	return NULL;
}

int Icons::_parseIconLocation(LPSTR szIconFile)
{
	LPSTR pszComma = strrchr(szIconFile, ',');

	if (!pszComma)
		return -1;

	*pszComma = 0;
	return atoi(pszComma + 1);
}

HICON Icons::_getIconFromPathEx(LPCSTR pszPath, int cxDesired, int cyDesired)
{
	char szIconFile[MAX_PATH];
	int nIconIndex = -1;
	LPCSTR pszExt;

    if(strlen(pszPath) > 0 && pszPath[0]) {
        if(stricmp(pszPath, ".none") == 0) {
            return NULL;
        }
        
        strncpy(szIconFile, pszPath, MAX_PATH);
        nIconIndex = _parseIconLocation(szIconFile);
        
        pszExt = strrchr(szIconFile, '.');
            
        if(pszExt && stricmp(pszExt, ".ico") == 0) {
            return (HICON) LoadImage(NULL, szIconFile, IMAGE_ICON, cxDesired, cyDesired, LR_LOADFROMFILE);
        }
        else if(nIconIndex >= 0) {
            HICON hIcon;
                
            ExtractIconEx(szIconFile, nIconIndex, (cxDesired <= 24 ? NULL : &hIcon), (cxDesired <= 24 ? &hIcon : NULL), 1);
                
            return hIcon;
        }
        else {
            SHFILEINFO shfi;
            char pszFileName[MAX_PATH];
            LPCSTR pszFile = szIconFile;
            UINT uFlags = NULL;
                
            if(strchr(szIconFile, '*')) {
                PathStripPath(szIconFile);
                uFlags |= SHGFI_USEFILEATTRIBUTES;
            }
            else if(!PathFileExists(szIconFile)) {
                FindExecutable(szIconFile, NULL, pszFileName);
                pszFile = pszFileName;
            }
                
            SHGetFileInfo(pszFile, 0, &shfi, sizeof(SHFILEINFO), SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0) | uFlags);
          
            return shfi.hIcon;
        }
    }
    else {
        // get generic folder icon
        SHFILEINFO shfi;
        
        SHGetFileInfo("*", FILE_ATTRIBUTE_DIRECTORY, &shfi, sizeof(SHFILEINFO), SHGFI_USEFILEATTRIBUTES | SHGFI_ICON | (cxDesired <= 24 ? SHGFI_SMALLICON : 0));
        
        return shfi.hIcon;
    }
}
